@@ -2,6 +2,8 @@
  * Copyright (C) 2004,2005,2006 Red Hat, Inc.
  * Authored by Christopher Aillon <caillon@redhat.com>
  *
+ * Copyright (C) 2008,2009 Guido Guenther <agx@sigxcup.org>
+ *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2, or (at your option)
@@ -36,6 +38,7 @@
 
 #include "krb5-auth-dialog.h"
 #include "krb5-auth-applet.h"
+#include "krb5-auth-pwdialog.h"
 #include "krb5-auth-gconf.h"
 #include "krb5-auth-dbus.h"
 
@@ -54,9 +57,10 @@ static krb5_timestamp canceled_creds_expiry;
 static gboolean canceled;
 static gboolean invalid_auth;
 static gboolean always_run;
+static gboolean is_online = TRUE;
 
-static int grab_credentials (Krb5AuthApplet* applet);
-static int ka_renew_credentials (Krb5AuthApplet* applet);
+static int grab_credentials (KaApplet* applet);
+static int ka_renew_credentials (KaApplet* applet);
 static gboolean ka_get_tgt_from_ccache (krb5_context context, krb5_creds *creds);
 
 /* YAY for different Kerberos implementations */
@@ -156,89 +160,70 @@ ka_krb5_cc_clear_mcred(krb5_creds* mcred)
 #endif
 }
 
+
 /* ***************************************************************** */
 /* ***************************************************************** */
 
 static gboolean
-credentials_expiring_real (Krb5AuthApplet* applet)
+credentials_expiring_real (KaApplet* applet)
 {
 	krb5_creds my_creds;
 	krb5_timestamp now;
 	gboolean retval = FALSE;
-	applet->renewable = FALSE;
 
+	ka_applet_set_tgt_renewable(applet, FALSE);
 	if (!ka_get_tgt_from_ccache (kcontext, &my_creds)) {
 		creds_expiry = 0;
 		retval = TRUE;
 		goto out;
 	}
 
+	/* copy principal from cache if any */
 	if (krb5_principal_compare (kcontext, my_creds.client, kprincipal)) {
 		krb5_free_principal(kcontext, kprincipal);
 		krb5_copy_principal(kcontext, my_creds.client, &kprincipal);
 	}
 	creds_expiry = my_creds.times.endtime;
 	if ((krb5_timeofday(kcontext, &now) == 0) &&
-	    (now + applet->pw_prompt_secs > my_creds.times.endtime))
+	    (now + ka_applet_get_pw_prompt_secs(applet) > my_creds.times.endtime))
 		retval = TRUE;
 
 	/* If our creds are expiring, determine whether they are renewable */
 	if (retval && get_cred_renewable(&my_creds) && my_creds.times.renew_till > now) {
-		applet->renewable = TRUE;
+		ka_applet_set_tgt_renewable(applet, TRUE);
 	}
 
 	krb5_free_cred_contents (kcontext, &my_creds);
 
 out:
-	ka_update_status(applet, creds_expiry);
+	ka_applet_update_status(applet, creds_expiry);
 	return retval;
 }
 
 
-static gchar* minutes_to_expiry_text (int minutes)
+/* time in seconds the tgt will be still valid */
+int
+ka_tgt_valid_seconds()
 {
-	gchar *expiry_text;
-	gchar *tmp;
+	krb5_timestamp now;
 
-	if (minutes > 0) {
-		expiry_text = g_strdup_printf (ngettext("Your credentials expire in %d minute",
-		                                        "Your credentials expire in %d minutes",
-		                                        minutes),
-		                               minutes);
-	} else {
-		expiry_text = g_strdup (_("Your credentials have expired"));
-		tmp = g_strdup_printf ("<span foreground=\"red\">%s</span>", expiry_text);
-		g_free (expiry_text);
-		expiry_text = tmp;
-	}
+	if (krb5_timeofday(kcontext, &now))
+		return 0;
 
-	return expiry_text;
+	return (creds_expiry - now);
 }
 
-
-static gboolean
-krb5_auth_dialog_wrong_label_update_expiry (GtkWidget* label)
+/* return credential cache filename, strip "FILE:" prefix if necessary */
+static const char*
+ka_ccache_filename (void)
 {
-	int minutes_left;
-	krb5_timestamp now;
-	gchar *expiry_text;
-	gchar *expiry_markup;
+	const gchar *ccache_name;
 
-	g_return_val_if_fail (label!= NULL, FALSE);
-
-	if (krb5_timeofday(kcontext, &now) != 0) {
-		return TRUE;
-	}
-
-	minutes_left = (creds_expiry - now) / 60;
-	expiry_text = minutes_to_expiry_text (minutes_left);
-
-	expiry_markup = g_strdup_printf ("<span size=\"smaller\" style=\"italic\">%s</span>", expiry_text);
-	gtk_label_set_markup (GTK_LABEL (label), expiry_markup);
-	g_free (expiry_text);
-	g_free (expiry_markup);
-
-	return TRUE;
+	ccache_name = krb5_cc_default_name (kcontext);
+	if (g_str_has_prefix (ccache_name, "FILE:"))
+		return &(ccache_name[5]);
+	else
+		return ccache_name;
 }
 
 
@@ -246,101 +231,30 @@ krb5_auth_dialog_wrong_label_update_expiry (GtkWidget* label)
 static gboolean
 krb5_auth_dialog_do_updates (gpointer data)
 {
-	Krb5AuthApplet* applet = (Krb5AuthApplet*)data;
-
-	g_return_val_if_fail (applet != NULL, FALSE);
+	KaApplet* applet = KA_APPLET(data);
+	KaPwDialog* pwdialog = ka_applet_get_pwdialog(applet);
 
+	g_return_val_if_fail (pwdialog != NULL, FALSE);
 	/* Update creds_expiry and close the applet if we got the creds by other means (e.g. kinit) */
-	if (!credentials_expiring_real(applet)) {
-		KA_DEBUG("PW Dialog persist: %d", applet->pw_dialog_persist);
-		if (!applet->pw_dialog_persist)
-			gtk_widget_hide(applet->pw_dialog);
-	}
+	if (!credentials_expiring_real(applet))
+		ka_pwdialog_hide(pwdialog, FALSE);
 
 	/* Update the expiry information in the dialog */
-	krb5_auth_dialog_wrong_label_update_expiry (applet->pw_wrong_label);
+	ka_pwdialog_status_update (pwdialog);
 	return TRUE;
 }
 
 
-static void
-krb5_auth_dialog_setup (Krb5AuthApplet *applet,
-                        const gchar *krb5prompt,
-                        gboolean hide_password)
-{
-	GtkWidget *entry;
-	GtkWidget *label;
-	gchar *wrong_text;
-	gchar *wrong_markup;
-	gchar *prompt;
-	int pw4len;
-
-
-	if (krb5prompt == NULL) {
-		prompt = g_strdup (_("Please enter your Kerberos password."));
-	} else {
-		/* Kerberos's prompts are a mess, and basically impossible to
-		 * translate.  There's basically no way short of doing a lot of
-		 * string parsing to translate them.  The most common prompt is
-		 * "Password for $uid:".  We special case that one at least.  We
-		 * cannot do any of the fancier strings (like challenges),
-		 * though. */
-		pw4len = strlen ("Password for ");
-		if (strncmp (krb5prompt, "Password for ", pw4len) == 0) {
-			gchar *uid = (gchar *) (krb5prompt + pw4len);
-			prompt = g_strdup_printf (_("Please enter the password for '%s'"), uid);
-		} else {
-			prompt = g_strdup (krb5prompt);
-		}
-	}
-
-	/* Clear the password entry field */
-	entry = glade_xml_get_widget (applet->pw_xml, "krb5_entry");
-	gtk_secure_entry_set_text (GTK_SECURE_ENTRY (entry), "");
-
-	/* Use the prompt label that krb5 provides us */
-	label = glade_xml_get_widget (applet->pw_xml, "krb5_message_label");
-	gtk_label_set_text (GTK_LABEL (label), prompt);
-
-	/* Add our extra message hints, if any */
-	wrong_text = NULL;
-
-	if (applet->pw_wrong_label) {
-		if (invalid_auth) {
-			wrong_text = g_strdup (_("The password you entered is invalid"));
-		} else {
-			krb5_timestamp now;
-			int minutes_left;
-
-			if (krb5_timeofday(kcontext, &now) == 0)
-				minutes_left = (creds_expiry - now) / 60;
-			else
-				minutes_left = 0;
-			wrong_text = minutes_to_expiry_text (minutes_left);
-		}
-	}
-
-	if (wrong_text) {
-		wrong_markup = g_strdup_printf ("<span size=\"smaller\" style=\"italic\">%s</span>", wrong_text);
-		gtk_label_set_markup (GTK_LABEL (applet->pw_wrong_label), wrong_markup);
-		g_free(wrong_text);
-		g_free(wrong_markup);
-	} else {
-		gtk_label_set_text (GTK_LABEL (applet->pw_wrong_label), "");
-	}
-	g_free (prompt);
-}
-
-
 static krb5_error_code
-auth_dialog_prompter (krb5_context ctx,
+auth_dialog_prompter (krb5_context ctx G_GNUC_UNUSED,
                       void *data,
-                      const char *name,
-                      const char *banner,
+                      const char *name G_GNUC_UNUSED,
+                      const char *banner G_GNUC_UNUSED,
                       int num_prompts,
                       krb5_prompt prompts[])
 {
-	Krb5AuthApplet* applet = (Krb5AuthApplet*)data;
+	KaApplet *applet = KA_APPLET(data);
+	KaPwDialog *pwdialog = ka_applet_get_pwdialog(applet);
 	krb5_error_code errcode;
 	int i;
 
@@ -354,27 +268,22 @@ auth_dialog_prompter (krb5_context ctx,
 		int response;
 		guint32 source_id;
 
-		GtkWidget *entry;
-
 		errcode = KRB5_LIBOS_CANTREADPWD;
 
-		krb5_auth_dialog_setup (applet, (gchar *) prompts[i].prompt, prompts[i].hidden);
-		entry = glade_xml_get_widget (applet->pw_xml, "krb5_entry");
-		gtk_widget_grab_focus (entry);
-
 		source_id = g_timeout_add_seconds (5, (GSourceFunc)krb5_auth_dialog_do_updates, applet);
-		response = gtk_dialog_run (GTK_DIALOG (applet->pw_dialog));
+		ka_pwdialog_setup (pwdialog, (gchar *) prompts[i].prompt, invalid_auth);
+		response = ka_pwdialog_run (pwdialog);
 		switch (response)
 		{
 			case GTK_RESPONSE_OK:
-				password = gtk_secure_entry_get_text (GTK_SECURE_ENTRY (entry));
+				password = ka_pwdialog_get_password(pwdialog);
 				password_len = strlen (password);
 				break;
+			case GTK_RESPONSE_DELETE_EVENT:
 			case GTK_RESPONSE_CANCEL:
 				canceled = TRUE;
 				break;
 			case GTK_RESPONSE_NONE:
-			case GTK_RESPONSE_DELETE_EVENT:
 				break;
 			default:
 				g_warning ("Unknown Response: %d", response);
@@ -394,14 +303,13 @@ auth_dialog_prompter (krb5_context ctx,
 		errcode = 0;
 	}
 cleanup:
+	ka_pwdialog_hide (pwdialog, TRUE);
 	/* Reset this, so we know the next time we get a TRUE value, it is accurate. */
-	gtk_widget_hide (applet->pw_dialog);
 	invalid_auth = FALSE;
 
 	return errcode;
 }
 
-static gboolean is_online = TRUE;
 
 #ifdef ENABLE_NETWORK_MANAGER
 static void
@@ -431,14 +339,15 @@ network_state_cb (libnm_glib_ctx *context,
 }
 #endif
 
+/* credentials expiring timer */
 static gboolean
 credentials_expiring (gpointer *data)
 {
 	int retval;
 	gboolean give_up;
-	Krb5AuthApplet* applet = (Krb5AuthApplet*) data;
+	KaApplet* applet = KA_APPLET(data);
 
-	KA_DEBUG("Checking expiry <%ds", applet->pw_prompt_secs);
+	KA_DEBUG("Checking expiry <%ds", ka_applet_get_pw_prompt_secs(applet));
 	if (credentials_expiring_real (applet) && is_online) {
 		KA_DEBUG("Expiry @ %ld", creds_expiry);
 
@@ -448,7 +357,7 @@ credentials_expiring (gpointer *data)
 		}
 
 		/* no popup when using a trayicon */
-		if (applet->show_trayicon)
+		if (ka_applet_get_show_trayicon(applet))
 			goto out;
 
 		give_up = canceled && (creds_expiry == canceled_creds_expiry);
@@ -465,13 +374,58 @@ credentials_expiring (gpointer *data)
 		}
 	}
 out:
-	ka_update_status(applet, creds_expiry);
+	ka_applet_update_status(applet, creds_expiry);
 	return TRUE;
 }
 
 
+/*
+ * set ticket options by looking at krb5.conf and gconf
+ */
 static void
-set_options_from_creds(const Krb5AuthApplet* applet,
+ka_set_ticket_options(KaApplet* applet, krb5_context context,
+		      krb5_get_init_creds_opt *out,
+		      const char* pk_userid, const char* pk_anchors)
+{
+	gboolean flag;
+#ifdef HAVE_KRB5_GET_INIT_CREDS_OPT_SET_DEFAULT_FLAGS
+	krb5_get_init_creds_opt_set_default_flags(context, PACKAGE,
+		krb5_principal_get_realm(context, kprincipal), out);
+#endif
+	g_object_get(applet, "tgt-forwardable", &flag, NULL);
+	if (flag)
+		krb5_get_init_creds_opt_set_forwardable(out, flag);
+	g_object_get(applet, "tgt-proxiable", &flag, NULL);
+	if (flag)
+		krb5_get_init_creds_opt_set_proxiable(out, flag);
+	g_object_get(applet, "tgt-renewable", &flag, NULL);
+	if (flag) {
+		krb5_deltat r = 3600*24*30; /* 1 month */
+		krb5_get_init_creds_opt_set_renew_life (out, r);
+	}
+
+#if ENABLE_PKINIT && HAVE_KRB5_GET_INIT_CREDS_OPT_SET_PA
+	/* pkinit optins for MIT Kerberos */
+	if (pk_userid && strlen(pk_userid)) {
+		KA_DEBUG("pkinit with '%s'", pk_userid);
+		krb5_get_init_creds_opt_set_pa(context, out,
+			"X509_user_identity", pk_userid);
+		if (pk_anchors && strlen(pk_anchors)) {
+			KA_DEBUG("pkinit anchors '%s'", pk_anchors);
+			krb5_get_init_creds_opt_set_pa(context, out,
+				"X509_anchors", pk_anchors);
+		}
+	}
+#endif /* HAVE_KRB5_GET_INIT_CREDS_OPT_SET_PA */
+}
+
+
+/*
+ * set ticket options
+ * by looking at krb5.conf, the passed in creds and gconf
+ */
+static void
+set_options_from_creds(const KaApplet* applet,
 		       krb5_context context,
 		       krb5_creds *in,
 		       krb5_get_init_creds_opt *out)
@@ -480,8 +434,8 @@ set_options_from_creds(const Krb5AuthApplet* applet,
 	int flag;
 
 #ifdef HAVE_KRB5_GET_INIT_CREDS_OPT_SET_DEFAULT_FLAGS
-	krb5_get_init_creds_opt_set_default_flags(kcontext, PACKAGE,
-		krb5_principal_get_realm(kcontext, kprincipal), out);
+	krb5_get_init_creds_opt_set_default_flags(context, PACKAGE,
+		krb5_principal_get_realm(context, kprincipal), out);
 #endif
 
 	flag = get_cred_forwardable(in) != 0;
@@ -496,7 +450,7 @@ set_options_from_creds(const Krb5AuthApplet* applet,
 						       renew_lifetime);
 	}
 	if (in->times.endtime >
-	    in->times.starttime + applet->pw_prompt_secs) {
+	    in->times.starttime + ka_applet_get_pw_prompt_secs(applet)) {
 		krb5_get_init_creds_opt_set_tkt_life(out,
 						     in->times.endtime -
 						     in->times.starttime);
@@ -506,27 +460,29 @@ set_options_from_creds(const Krb5AuthApplet* applet,
 }
 
 
+#if ENABLE_PKINIT && HAVE_KRB5_GET_INIT_CREDS_OPT_SET_PKINIT
 static krb5_error_code
-ka_auth_pkinit(Krb5AuthApplet* applet, krb5_creds* creds)
+ka_auth_heimdal_pkinit(KaApplet* applet, krb5_creds* creds,
+                       const char* pk_userid, const char* pk_anchors)
 {
-#ifdef ENABLE_PKINIT
 	krb5_get_init_creds_opt *opts = NULL;
 	krb5_error_code retval;
+	const char* pkinit_anchors = NULL;
 
-	KA_DEBUG("pkinit with %s", applet->pk_userid);
+	KA_DEBUG("pkinit with '%s'", pk_userid);
+	if (pk_anchors && strlen (pk_anchors)) {
+		pkinit_anchors = pk_anchors;
+		KA_DEBUG("pkinit anchors '%s'", pkinit_anchors);
+	}
 
-	if (!applet->pk_userid)
-		return 0;
-
-	retval = krb5_get_init_creds_opt_alloc (kcontext, &opts);
-	if (retval)
+	if ((retval = krb5_get_init_creds_opt_alloc (kcontext, &opts)))
 		goto out;
-	set_options_from_creds (applet, kcontext, creds, opts);
 
+	ka_set_ticket_options (applet, kcontext, opts, NULL, NULL);
 	retval = krb5_get_init_creds_opt_set_pkinit(kcontext, opts,
 						    kprincipal,
-						    applet->pk_userid,
-						    NULL, /* x509 anchors */
+						    pk_userid,
+						    pkinit_anchors,
 						    NULL,
 						    NULL,
 						    0,	  /* pk_use_enc_key */
@@ -541,52 +497,148 @@ ka_auth_pkinit(Krb5AuthApplet* applet, krb5_creds* creds)
 	                                      NULL, auth_dialog_prompter, applet,
 		                              0, NULL, opts);
 out:
-	krb5_get_init_creds_opt_free(kcontext, opts);
+	if (opts)
+		krb5_get_init_creds_opt_free(kcontext, opts);
 	return retval;
-#else  /* ENABLE_PKINIT */
-	return 0;
+}
 #endif /* ! ENABLE_PKINIT */
+
+static krb5_error_code
+ka_auth_password(KaApplet* applet, krb5_creds* creds,
+                 const char* pk_userid, const char* pk_anchors)
+{
+	krb5_error_code retval;
+	krb5_get_init_creds_opt *opts = NULL;
+
+	if ((retval = krb5_get_init_creds_opt_alloc (kcontext, &opts)))
+		goto out;
+	ka_set_ticket_options (applet, kcontext, opts,
+	                       pk_userid, pk_anchors);
+
+	retval = krb5_get_init_creds_password(kcontext, creds, kprincipal,
+					      NULL, auth_dialog_prompter, applet,
+					      0, NULL, opts);
+out:
+	if (opts)
+		krb5_get_init_creds_opt_free(kcontext, opts);
+	return retval;
+}
+
+static krb5_error_code
+ka_parse_name(KaApplet* applet, krb5_context krbcontext, krb5_principal* kprinc)
+{
+	krb5_error_code ret;
+	gchar *principal = NULL;
+
+	if (*kprinc != NULL)
+		krb5_free_principal(krbcontext, *kprinc);
+
+	g_object_get(applet, "principal", &principal, NULL);
+	ret = krb5_parse_name(krbcontext, principal, kprinc);
+
+	g_free(principal);
+	return ret;
+}
+
+
+static void
+ccache_changed_cb (GFileMonitor *monitor G_GNUC_UNUSED,
+                   GFile *file,
+                   GFile *other_file G_GNUC_UNUSED,
+                   GFileMonitorEvent event_type,
+                   gpointer data)
+{
+	KaApplet *applet = KA_APPLET(data);
+	gchar *ccache_name = g_file_get_path(file);
+
+	switch (event_type) {
+		case G_FILE_MONITOR_EVENT_DELETED:
+		case G_FILE_MONITOR_EVENT_CREATED:
+		case G_FILE_MONITOR_EVENT_CHANGED:
+			KA_DEBUG ("%s changed", ccache_name);
+			credentials_expiring ((gpointer)applet);
+			break;
+		default:
+			KA_DEBUG ("%s unhandled event: %d", ccache_name, event_type);
+	}
+	g_free (ccache_name);
+}
+
+
+static gboolean
+monitor_ccache(KaApplet* applet)
+{
+	const gchar *ccache_name;
+	GFile *ccache;
+	GFileMonitor *monitor;
+	GError *err = NULL;
+	gboolean ret = FALSE;
+
+	ccache_name = ka_ccache_filename ();
+	g_return_val_if_fail (ccache_name != NULL, FALSE);
+
+	ccache = g_file_new_for_path (ccache_name);
+	monitor = g_file_monitor_file (ccache, G_FILE_MONITOR_NONE, NULL, &err);
+	g_assert ((!monitor && err) || (monitor && !err));
+	if (!monitor) {
+		/* cache disappeared? */
+		if (err->code == G_FILE_ERROR_NOENT)
+			credentials_expiring ((gpointer)applet);
+		else
+			g_warning ("Failed to monitor %s: %s", ccache_name, err->message);
+		goto out;
+	} else {
+		/* g_file_monitor_set_rate_limit(monitor, 10*1000); */
+		g_signal_connect (monitor, "changed", G_CALLBACK (ccache_changed_cb), applet);
+		KA_DEBUG ("Monitoring %s", ccache_name);
+		ret = TRUE;
+	}
+out:
+	g_object_unref (ccache);
+	if (err)
+		g_error_free (err);
+	return ret;
 }
 
 
 /* grab credentials interactively */
 static int
-grab_credentials (Krb5AuthApplet* applet)
+grab_credentials (KaApplet* applet)
 {
-	krb5_error_code retval;
+	krb5_error_code retval = KRB5_KDC_UNREACH;
 	krb5_creds my_creds;
 	krb5_ccache ccache;
-	krb5_get_init_creds_opt *opt = NULL;
+	gchar *pk_userid = NULL;
+	gchar *pk_anchors = NULL;
+	gboolean pw_auth = TRUE;
 
 	memset(&my_creds, 0, sizeof(my_creds));
 
-	if (kprincipal == NULL) {
-		retval = krb5_parse_name(kcontext, applet->principal,
-					 &kprincipal);
-		if (retval) {
-			return retval;
-		}
-	}
+	retval = ka_parse_name(applet, kcontext, &kprincipal);
+	if (retval)
+		goto out2;
 
 	retval = krb5_cc_default (kcontext, &ccache);
 	if (retval)
-		return retval;
+		goto out2;
 
-#if ENABLE_PKINIT
-	if (applet->pk_userid) { /* try pkinit */
-#else
-	if (0) {
-#endif
-		retval = ka_auth_pkinit(applet, &my_creds);
-	} else {
-		retval = krb5_get_init_creds_opt_alloc (kcontext, &opt);
-		if (retval)
-			goto out;
-		set_options_from_creds (applet, kcontext, &my_creds, opt);
-		retval = krb5_get_init_creds_password(kcontext, &my_creds, kprincipal,
-						      NULL, auth_dialog_prompter, applet,
-						      0, NULL, opt);
+	g_object_get(applet, "pk-userid", &pk_userid,
+	                     "pk-anchors", &pk_anchors,
+	                     NULL);
+#if ENABLE_PKINIT && HAVE_HX509_ERR_H && HAVE_KRB5_GET_INIT_CREDS_OPT_SET_PKINIT
+	/* pk_userid set: try pkinit */
+	if (pk_userid && strlen(pk_userid)) {
+		retval = ka_auth_heimdal_pkinit(applet, &my_creds,
+		                                pk_userid, pk_anchors);
+		/* other error than: "no token found" - no need to try password auth: */
+		if (retval != HX509_PKCS11_NO_TOKEN && retval != HX509_PKCS11_NO_SLOT)
+			pw_auth = FALSE;
 	}
+#endif /* ENABLE_PKINIT */
+	if (pw_auth)
+		retval = ka_auth_password(applet, &my_creds,
+		                          pk_userid, pk_anchors);
+
 	creds_expiry = my_creds.times.endtime;
 	if (canceled)
 		canceled_creds_expiry = creds_expiry;
@@ -596,10 +648,9 @@ grab_credentials (Krb5AuthApplet* applet)
 			case KRB5KRB_AP_ERR_BAD_INTEGRITY:
 #ifdef HAVE_HX509_ERR_H
 			case HX509_PKCS11_LOGIN:
-#endif
-				/* Invalid password/pin, try again. */
+#endif 				/* Invalid password/pin, try again. */
 				invalid_auth = TRUE;
-				goto out;
+				break;
 			default:
 				KA_DEBUG("Auth failed with %d: %s", retval,
 				         get_error_message(kcontext, retval));
@@ -614,19 +665,17 @@ grab_credentials (Krb5AuthApplet* applet)
 	retval = krb5_cc_store_cred(kcontext, ccache, &my_creds);
 	if (retval)
 		goto out;
-
 out:
-	if (opt)
-		krb5_get_init_creds_opt_free(kcontext, opt);
 	krb5_free_cred_contents (kcontext, &my_creds);
 	krb5_cc_close (kcontext, ccache);
-
+out2:
+	g_free(pk_userid);
 	return retval;
 }
 
 /* try to renew the credentials noninteractively */
 static int
-ka_renew_credentials (Krb5AuthApplet* applet)
+ka_renew_credentials (KaApplet* applet)
 {
 	krb5_error_code retval;
 	krb5_creds my_creds;
@@ -634,8 +683,7 @@ ka_renew_credentials (Krb5AuthApplet* applet)
 	krb5_get_init_creds_opt opts;
 
 	if (kprincipal == NULL) {
-		retval = krb5_parse_name(kcontext, applet->principal,
-					 &kprincipal);
+		retval = ka_parse_name(applet, kcontext, &kprincipal);
 		if (retval)
 			return retval;
 	}
@@ -653,7 +701,7 @@ ka_renew_credentials (Krb5AuthApplet* applet)
 	krb5_get_init_creds_opt_init (&opts);
 	set_options_from_creds (applet, kcontext, &my_creds, &opts);
 
-	if (applet->renewable) {
+	if (ka_applet_get_tgt_renewable(applet)) {
 		retval = get_renewed_creds (kcontext, &my_creds, kprincipal, ccache, NULL);
 		if (retval)
 			goto out;
@@ -718,7 +766,7 @@ out:
 }
 
 static gboolean
-using_krb5()
+using_krb5(void)
 {
 	krb5_error_code err;
 	gboolean have_tgt = FALSE;
@@ -733,15 +781,15 @@ using_krb5()
 		krb5_copy_principal(kcontext, creds.client, &kprincipal);
 		krb5_free_cred_contents (kcontext, &creds);
 	}
-
 	return have_tgt;
 }
 
 
 void
-ka_destroy_cache (GtkMenuItem  *menuitem, gpointer data)
+ka_destroy_cache (GtkMenuItem  *menuitem G_GNUC_UNUSED,
+		  gpointer data)
 {
-	Krb5AuthApplet* applet = (Krb5AuthApplet*) data;
+	KaApplet *applet = KA_APPLET(data);
 	krb5_ccache  ccache;
 	const char* cache;
 	krb5_error_code ret;
@@ -768,42 +816,133 @@ ka_error_dialog(int err)
 }
 
 
-/* this is done on leftclick, update the tooltip immediately */
-void
-ka_grab_credentials (Krb5AuthApplet* applet)
+/*
+ * check if we have valid credentials for the requested principal - if not, grab them
+ * principal: requested principal - if empty use default
+ */
+gboolean
+ka_check_credentials (KaApplet *applet, const char* newprincipal)
+{
+	gboolean success = FALSE;
+	int retval;
+	char* principal;
+
+	g_object_get(applet, "principal", &principal, NULL);
+
+	if (strlen(newprincipal)) {
+		krb5_principal knewprinc;
+
+		/* no ticket cache: is requested princ the one from our config? */
+		if (!kprincipal && g_strcmp0(principal, newprincipal)) {
+			KA_DEBUG("Requested principal %s not %s", principal, newprincipal);
+			goto out;
+		}
+
+		/* ticket cache: check if the requested principal is the one we have */
+		retval = krb5_parse_name(kcontext, newprincipal, &knewprinc);
+		if (retval) {
+			g_warning ("Cannot parse principal '%s'", newprincipal);
+			goto out;
+		}
+		if (kprincipal && !krb5_principal_compare (kcontext, kprincipal, knewprinc)) {
+			KA_DEBUG("Current Principal '%s' not '%s'", principal, newprincipal);
+		        krb5_free_principal(kcontext, knewprinc);
+			goto out;
+		}
+		krb5_free_principal(kcontext, knewprinc);
+	}
+
+	if (credentials_expiring_real (applet)) {
+		if (!is_online)
+			success = FALSE;
+		else
+			success = ka_grab_credentials (applet);
+	} else
+		success = TRUE;
+out:
+	g_free (principal);
+	return success;
+}
+
+
+/* initiate grabbing of credentials (e.g. on leftclick of tray icon) */
+gboolean
+ka_grab_credentials (KaApplet* applet)
 {
 	int retval;
-	gboolean retry;
+	int success = FALSE;
+	KaPwDialog *pwdialog = ka_applet_get_pwdialog(applet);
 
-	applet->pw_dialog_persist = TRUE;
+	ka_pwdialog_set_persist(pwdialog, TRUE);
 	do {
-		retry = TRUE;
 		retval = grab_credentials (applet);
 		if (invalid_auth)
 			continue;
-		switch (retval) {
-		    case 0: /* success */
-		    case KRB5_LIBOS_PWDINTR:     /* canceled (heimdal) */
-		    case KRB5_LIBOS_CANTREADPWD: /* canceled (mit) */
-			    retry = FALSE;
-			    break;
-		    case KRB5KDC_ERR_C_PRINCIPAL_UNKNOWN:
-		    default:
-			    ka_error_dialog(retval);
-			    retry = FALSE;
-			    break;
+		if (canceled)
+			break;
+		if (retval) {
+			ka_error_dialog(retval);
+			break;
+		} else {
+			success = TRUE;
+			break;
 		}
-	} while(retry);
+	} while(TRUE);
 
-	applet->pw_dialog_persist = FALSE;
+	ka_pwdialog_set_persist(pwdialog, FALSE);
 	credentials_expiring_real(applet);
+
+	return success;
+}
+
+
+static void
+ka_secmem_init (void)
+{
+	/* Initialize secure memory.  1 is too small, so the default size
+	will be used.  */
+	secmem_init (1);
+	secmem_set_flags (SECMEM_WARN);
+	drop_privs ();
+
+	if (atexit (secmem_term))
+		g_error("Couln't register atexit handler");
+}
+
+
+static gboolean
+ka_nm_init(void)
+{
+#ifdef ENABLE_NETWORK_MANAGER
+	libnm_glib_ctx *nm_context;
+	guint32 nm_callback_id;
+
+	nm_context = libnm_glib_init ();
+	if (!nm_context) {
+		g_warning ("Could not initialize libnm_glib");
+	} else {
+		nm_callback_id = libnm_glib_register_callback (nm_context, network_state_cb, &is_online, NULL);
+		if (nm_callback_id == 0) {
+			libnm_glib_shutdown (nm_context);
+			nm_context = NULL;
+
+			g_warning ("Could not connect to NetworkManager, connection status will not be managed!");
+		}
+	}
+#endif /* ENABLE_NETWORK_MANAGER */
+	return TRUE;
 }
 
 
 static GtkWidget*
-ka_create_gtk_secure_entry (GladeXML *xml, gchar *func_name, gchar *name,
-				gchar *s1, gchar *s2, gint i1, gint i2,
-				gpointer user_data)
+ka_create_gtk_secure_entry (GladeXML *xml G_GNUC_UNUSED,
+			    gchar *func_name G_GNUC_UNUSED,
+			    gchar *name,
+			    gchar *s1 G_GNUC_UNUSED,
+			    gchar *s2 G_GNUC_UNUSED,
+			    gint i1 G_GNUC_UNUSED,
+			    gint i2 G_GNUC_UNUSED,
+			    gpointer user_data G_GNUC_UNUSED)
 {
 	GtkWidget* entry = NULL;
 
@@ -818,26 +957,13 @@ ka_create_gtk_secure_entry (GladeXML *xml, gchar *func_name, gchar *name,
 }
 
 
-static void
-ka_secmem_init ()
-{
-	/* Initialize secure memory.  1 is too small, so the default size
-	will be used.  */
-	secmem_init (1);
-	secmem_set_flags (SECMEM_WARN);
-	drop_privs ();
-
-	if (atexit (secmem_term))
-		g_error("Couln't register atexit handler");
-}
-
-
 int
 main (int argc, char *argv[])
 {
-	Krb5AuthApplet *applet;
+	KaApplet *applet;
 	GOptionContext *context;
 	GError *error = NULL;
+	GladeXML *xml;
 
 	guint status = 0;
 	gboolean run_auto = FALSE, run_always = FALSE;
@@ -851,10 +977,6 @@ main (int argc, char *argv[])
   		{ NULL, 0, 0, G_OPTION_ARG_NONE, NULL, NULL, NULL }
 	};
 
-#ifdef ENABLE_NETWORK_MANAGER
-	libnm_glib_ctx *nm_context;
-	guint32 nm_callback_id;	
-#endif
 	context = g_option_context_new ("- Kerberos 5 credential checking");
 	g_option_context_add_main_entries (context, options, NULL);
 	g_option_context_add_group (context, gtk_get_option_group (TRUE));
@@ -878,40 +1000,25 @@ main (int argc, char *argv[])
 		always_run = TRUE;
 	}
 	if (using_krb5 () || always_run) {
-		applet = ka_create_applet ();
+		g_set_application_name (_("Network Authentication"));
+
+		glade_set_custom_handler (&ka_create_gtk_secure_entry, NULL);
+		xml = glade_xml_new (KA_DATA_DIR G_DIR_SEPARATOR_S
+				     PACKAGE ".glade", NULL, NULL);
+		applet = ka_applet_create (xml);
 		if (!applet)
 			return 1;
 		if (!ka_gconf_init (applet, argc, argv))
 			return 1;
-
-		/* setup the pw dialog */
-		glade_set_custom_handler (&ka_create_gtk_secure_entry, NULL);
-		applet->pw_xml = glade_xml_new (GLADEDIR "krb5-auth-dialog.glade", NULL, NULL);
-		applet->pw_wrong_label = glade_xml_get_widget (applet->pw_xml, "krb5_wrong_label");
-		applet->pw_dialog = glade_xml_get_widget (applet->pw_xml, "krb5_dialog");
-
-		g_set_application_name (_("Network Authentication"));
-
-#ifdef ENABLE_NETWORK_MANAGER
-		nm_context = libnm_glib_init ();
-		if (!nm_context) {
-			g_warning ("Could not initialize libnm_glib");
-		} else {
-			nm_callback_id = libnm_glib_register_callback (nm_context, network_state_cb, &is_online, NULL);
-			if (nm_callback_id == 0) {
-				libnm_glib_shutdown (nm_context);
-				nm_context = NULL;
-
-				g_warning ("Could not connect to NetworkManager, connection status will not be managed!");
-			}
-		}
-#endif /* ENABLE_NETWORK_MANAGER */
+		ka_nm_init();
 
 		if (credentials_expiring ((gpointer)applet)) {
 			g_timeout_add_seconds (CREDENTIAL_CHECK_INTERVAL, (GSourceFunc)credentials_expiring, applet);
+			monitor_ccache (applet);
 		}
+		ka_dbus_service(applet);
 		gtk_main ();
+		g_object_unref(xml);
 	}
-
 	return 0;
 }
